# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Collectors serves as a wrapper on top garf query."""

import os
import pathlib
from typing import Union

import pydantic
import smart_open
import yaml


class Collector(pydantic.BaseModel):
  """Stores API query with meta information.

  Attributes:
    query: Text of a query for an API.
    title: Identifier of a query.
    suffix: Optional element to be added to Prometheus metric.
  """

  query: str
  title: str
  suffix: str = ''


def load_collector_data(
  path_to_definitions: Union[str, os.PathLike[str], pathlib.Path],
) -> list[Collector]:
  """Loads collectors data from file or folder.

  Args:
    path_to_definitions: Local path to file / folder with collector definitions.

  Returns:
    Loaded collector definitions.
  """
  if isinstance(path_to_definitions, str):
    path_to_definitions = pathlib.Path(path_to_definitions)
  results = []
  if path_to_definitions.is_file():
    with smart_open.open(path_to_definitions, 'r', encoding='utf-8') as f:
      data = yaml.safe_load(f)
      for entry in data:
        results.append(Collector(**entry))
  else:
    for file in path_to_definitions.iterdir():
      if file.suffix == '.yaml':
        with smart_open.open(file, 'r', encoding='utf-8') as f:
          data = yaml.safe_load(f)
          for entry in data:
            results.append(Collector(**entry))
  return results



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=C0330, g-bad-import-order, g-multiple-import

"""Handles exporting all metrics based on a request."""

from __future__ import annotations

import logging
import os
from time import time
from typing import Literal, Optional, Union

import garf.core
import garf.exporter
import pydantic
from garf.executors import fetchers
from garf.exporter import collector
from garf.exporter.telemetry import tracer

logger = logging.getLogger(__name__)


class GarfExporterRuntimeOptions(pydantic.BaseModel):
  """Options to finetune exporting process.

  Attributes:
    expose_type: Type of exposition (http or pushgateway).
    host: Address of expose endpoint.
    port: Port of expose endpoint.
    iterations: Optional number of iterations to perform.
    delay_minutes: Delay between exports.
    namespace: Prefix for all metrics exposed to Prometheus.
    job_name: Job name attached to each metric.
    fetching_timeout:
      Period to abort fetching if not data from API returned.
    max_workers: Maximum number of parallel fetching from an API.
  """

  host: str = '0.0.0.0'
  port: int = 8000
  expose_type: Literal['http', 'pushgateway'] = 'http'
  iterations: Optional[int] = None
  delay_minutes: int = 15
  namespace: str = 'garf'
  job_name: str = 'garf_exporter'
  fetching_timeout: int = 120
  max_workers: Optional[int] = None

  @property
  def address(self) -> str:
    return f'{self.host}:{self.port}'


class GarfExporterRequest(pydantic.BaseModel):
  """Request to API to perform export to Prometheus.

  Attributes:
    source: Type of API to get data from.
    source_parameters: Optional parameters to configure the API connection.
    collectors_config: Path to YAML file with collector definitions.
    query_parameters: Optional parameters to refine queries in collectors.
    runtime_options: Options to finetune exporting process.
  """

  source: str
  source_parameters: dict[str, Union[str, int]] = pydantic.Field(
    default_factory=dict
  )
  collectors_config: Union[os.PathLike[str], str, None] = None
  query_parameters: Optional[garf.core.query_editor.GarfQueryParameters] = None
  runtime_options: GarfExporterRuntimeOptions = GarfExporterRuntimeOptions()
  collectors: list[collector.Collector] = pydantic.Field(default_factory=list)


class GarfExporterService:
  """Responsible for getting data from API and exposing it to Prometheus.

  Attributes:
    alias: Type of fetcher used to get data from an API.
    fetcher: Initialized ApiReportFetcher to perform fetching from an API.
  """

  def __init__(
    self,
    alias: str,
    source_parameters: Optional[dict[str, Union[str, int]]] = None,
  ) -> None:
    """Initializes GarfExporterService."""
    self.alias = alias
    self.source_parameters = source_parameters or {}
    self._report_fetcher = None

  @property
  def fetcher(self) -> garf.core.ApiReportFetcher:
    """Initialized ApiReportFetcher for fetching data from an API."""
    if self._report_fetcher:
      return self._report_fetcher
    fetcher = fetchers.get_report_fetcher(self.alias)
    self._report_fetcher = fetcher(**self.source_parameters)
    return self._report_fetcher

  @tracer.start_as_current_span('generate_metrics')
  def generate_metrics(
    self,
    request: GarfExporterRequest,
    exporter: garf.exporter.GarfExporter,
  ) -> None:
    """Generates metrics based on API request.

    Args:
      request: Complete request to fetch and expose data.
      exporter: Initialized GarfExporter.
    """
    logger.info('Beginning export')
    start_export_time = time()
    exporter.export_started.set(start_export_time)
    collectors = request.collectors or collector.load_collector_data(
      request.collectors_config
    )
    for col in collectors:
      logger.info('Exporting from collector: %s', col.title)
      start = time()
      report = self.fetcher.fetch(
        col.query, args=request.query_parameters, **self.source_parameters
      )
      end = time()
      exporter.report_fetcher_gauge.labels(collector=col.title).set(end - start)
      exporter.export(
        report=report,
        suffix=col.suffix,
        collector=col.title,
      )
    logger.info('Export completed')
    end_export_time = time()
    exporter.export_completed.set(end_export_time)
    exporter.total_export_time_gauge.set(end_export_time - start_export_time)
    exporter.delay_gauge.set(request.runtime_options.delay_minutes * 60)



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=C0330, g-bad-import-order, g-multiple-import

"""GarfExporter is responsible for preparing data for Prometheus."""

from __future__ import annotations

import logging
import time
from collections import abc
from collections.abc import Sequence

import garf.core
import prometheus_client
from garf.exporter.telemetry import tracer

logger = logging.getLogger(__name__)


class GarfExporter:
  """Exposes reports from Ads API in Prometheus format.

  Attributes:
    namespace: Global prefix for all Prometheus metrics.
    job_name: Name of export job in Prometheus.
    expose_metrics_with_zero_values: Whether to send zero metrics.
  """

  def __init__(
    self,
    namespace: str = 'garf',
    job_name: str = 'garf_exporter',
    expose_metrics_with_zero_values: bool = False,
  ) -> None:
    """Initializes GarfExporter to serve metrics.

    Args:
      namespace: Global prefix for all Prometheus metrics.
      job_name: Name of export job in Prometheus.
      expose_metrics_with_zero_values: Whether to send zero metrics.
    """
    self.namespace = namespace
    self.job_name = job_name
    self.expose_metrics_with_zero_values = expose_metrics_with_zero_values
    self.registry: prometheus_client.CollectorRegistry = (
      prometheus_client.CollectorRegistry()
    )

  @property
  def export_started(self) -> prometheus_client.Gauge:
    """Gauge for tracking start of collectors export."""
    return self._define_gauge('export_started_seconds', suffix='Remove')

  @property
  def export_completed(self) -> prometheus_client.Gauge:
    """Gauge for tracking end of collectors export."""
    return self._define_gauge('export_completed_seconds', suffix='Remove')

  @property
  def total_export_time_gauge(self) -> prometheus_client.Gauge:
    """Gauge for tracking exports in seconds."""
    return self._define_gauge('exporting_seconds', suffix='Remove')

  @property
  def report_fetcher_gauge(self) -> prometheus_client.Gauge:
    """Gauge for tracking report fetching for account in seconds."""
    return self._define_gauge(
      name='report_fetching_seconds',
      suffix='Remove',
      labelnames=('collector',),
    )

  @property
  def delay_gauge(self) -> prometheus_client.Gauge:
    """Gauge for exposing delay between exports."""
    return self._define_gauge('delay_seconds', suffix='Remove')

  def reset_registry(self) -> None:
    """Removes all metrics from registry before export."""
    self.registry._collector_to_names.clear()
    self.registry._names_to_collectors.clear()

  @tracer.start_as_current_span('export')
  def export(
    self,
    report: garf.core.GarfReport,
    suffix: str = '',
    collector: str | None = None,
  ) -> None:
    """Exports data from report into the format consumable by Prometheus.

    Iterates over each row or report and creates gauges (metrics with labels
    attached to them) which are added to the registry.

    Args:
      report: Report with API data.
      suffix: Common identifier to be added to a series of metrics.
      collector: Name of one of GarfExporter collectors attached to report.
    """
    if not report:
      return
    start = time.time()
    export_time_gauge = self._define_gauge(
      name='query_export_time_seconds',
      suffix='Remove',
      labelnames=('collector',),
    )
    api_requests_counter = self._define_counter(name='api_requests_count')
    metrics = self._define_metrics(report.query_specification, suffix)
    labels = self._define_labels(report.query_specification)
    for row in report:
      label_values = []
      for label in labels:
        if isinstance(row.get(label), abc.MutableSequence):
          label_value = ','.join([str(r) for r in row.get(label)])
        else:
          label_value = row.get(label)
        label_values.append(label_value)
      for name, metric in metrics.items():
        if (
          metric_value := getattr(row, name)
          or self.expose_metrics_with_zero_values
        ) and not isinstance(metric_value, str):
          metric.labels(*label_values).set(metric_value)
    end = time.time()
    export_time_gauge.labels(collector=collector).set(end - start)
    api_requests_counter.inc()
    self.registry.collect()

  def _define_metrics(
    self,
    query_specification: garf.core.query_editor.QuerySpecification,
    suffix: str,
  ) -> dict[str, prometheus_client.Gauge]:
    """Defines metrics to be exposed Prometheus.

    Metrics are defined based on query_specification of report that needs to
    be exposed. It takes into account both virtual and non-virtual columns.

    Args:
      query_specification:
        QuerySpecification that contains all information about the query.
      suffix: Common identifier to be added to a series of metrics.

    Returns:
      Mapping between metrics alias in report and Gauge.
    """
    metrics = {}
    labels = self._define_labels(query_specification)
    non_virtual_columns = self._get_non_virtual_columns(query_specification)
    for column, field in zip(non_virtual_columns, query_specification.fields):
      if 'metric' in field or 'metric' in column:
        metrics[column] = self._define_gauge(column, suffix, labels)
    if virtual_columns := query_specification.virtual_columns:
      for column, field in virtual_columns.items():
        metrics[column] = self._define_gauge(column, suffix, labels)
    logger.debug('metrics: %s', metrics)
    return metrics

  def _define_labels(
    self, query_specification: garf.core.query_editor.QuerySpecification
  ) -> list[str]:
    """Defines names of labels to be attached to metrics.

    Label names are build based on column names of the report. Later on each
    label name gets its own value (i.e. customer_id=1, campaign_type=DISPLAY).

    Args:
      query_specification:
        QuerySpecification that contains all information about the query.
      suffix: Common identifier to be added to a series of metrics.

    Returns:
      All possible labels names that can be attached to metrics.
    """
    labelnames = []
    non_virtual_columns = self._get_non_virtual_columns(query_specification)
    for column, field in zip(non_virtual_columns, query_specification.fields):
      if 'metric' not in field and 'metric' not in column:
        labelnames.append(str(column))
    logger.debug('labelnames: %s', labelnames)
    return labelnames

  def _define_gauge(
    self,
    name: str,
    suffix: str,
    labelnames: Sequence[str] = (),
  ) -> prometheus_client.Gauge:
    """Defines Gauge metric to be created in Prometheus and add labels to it.

    Gauge has the following structure '<namespace>_<suffix>_<name>' and might
    look like this `googleads_disappoved_ads_count` meaning that it comes from
    `googleads` namespace (usually common for all metrics), `disapproved_ads`
    signifies that one or several metrics are coming from a single data fetch
    and usually grouped logically, while `count` represent the metric itself.

    Args:
      name: Name of the metric to be exposed to Prometheus (without prefix).
      suffix: Common identifier to be added to a series of metrics.
      labelnames: Dimensions attached to metric (i.e. ad_group_id, account).

    Returns:
      An instance of Counter that associated with registry.
    """
    if suffix and suffix != 'Remove':
      gauge_name = f'{self.namespace}_{suffix}_{name}'
    else:
      gauge_name = f'{self.namespace}_{name}'
    if gauge_name in self.registry._names_to_collectors:
      return self.registry._names_to_collectors.get(gauge_name)
    return prometheus_client.Gauge(
      name=gauge_name,
      documentation=name,
      labelnames=labelnames,
      registry=self.registry,
    )

  def _define_counter(self, name: str) -> prometheus_client.Counter:
    """Define Counter metric based on provided name.

    Args:
      name: Name of the metric to be exposed to Prometheus (without prefix).

    Returns:
      An instance of Counter that associated with registry.
    """
    counter_name = f'{self.namespace}_{name}'
    if counter_name in self.registry._names_to_collectors:
      return self.registry._names_to_collectors.get(counter_name)
    return prometheus_client.Counter(
      name=counter_name, documentation=name, registry=self.registry
    )

  def _get_non_virtual_columns(
    self, query_specification: garf.core.query_editor.QuerySpecification
  ) -> list[str]:
    """Returns all non-virtual columns from query.

    Virtual columns have special handling during the export so they need
    to be removed.

    Args:
      query_specification:
        QuerySpecification that contains all information about the query.

    Returns:
      All columns from the query that are not virtual.
    """
    return [
      column
      for column in query_specification.column_names
      if column not in query_specification.virtual_columns
    ]

  def __str__(self) -> str:  # noqa: D105
    return f'GarfExporter(namespace={self.namespace}, job_name={self.job_name})'



# Copyright 2025 Google LLC
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=C0330, g-bad-import-order, g-multiple-import

"""Entrypoint for running GarfExporter.

Defines GarfExporter collectors, fetches data from API
and exposes them to Prometheus.
"""

from __future__ import annotations

import asyncio
import contextlib
import datetime
import logging

import fastapi
import garf.exporter
import prometheus_client
import requests
import typer
import uvicorn
from garf.executors.entrypoints import utils as garf_utils
from garf.exporter import exporter_service
from garf.exporter.entrypoints.tracer import initialize_tracer
from opentelemetry.instrumentation.fastapi import FastAPIInstrumentor
from typing_extensions import Annotated

typer_app = typer.Typer()


class GarfExporterError(Exception):
  """Base class for GarfExporter errors."""


def healthcheck(host: str, port: int) -> bool:
  """Validates that the GarfExporter export happened recently.

  Healthcheck compares the time passed since the last successful export with
  the delay between exports. If this delta if greater than 1.5 check is failed.

  Args:
    host: Hostname garf-exporter http server (i.e. localhost).
    port: Port garf-exporter http server is running (i.e. 8000).


  Returns:
    Whether or not the check is successful.
  """
  try:
    res = requests.get(f'http://{host}:{port}/metrics/').text.split('\n')
  except requests.exceptions.ConnectionError:
    return False
  last_exported = [r for r in res if 'export_completed_seconds 1' in r][
    0
  ].split(' ')[1]
  delay = None
  for result in [r for r in res if 'delay_seconds' in r]:
    _, *value = result.split(' ', maxsplit=2)
    with contextlib.suppress(ValueError):
      delay = float(value[0])
  if not delay:
    return False

  max_allowed_delta = 1.5
  is_lagged_export = (
    datetime.datetime.now().timestamp() - float(last_exported)
  ) > (max_allowed_delta * delay)

  return not is_lagged_export


initialize_tracer()
app = fastapi.FastAPI(debug=False)
exporter = garf.exporter.GarfExporter()
metrics_app = prometheus_client.make_asgi_app(registry=exporter.registry)
app.mount('/metrics', metrics_app)
FastAPIInstrumentor.instrument_app(app)


async def start_metric_generation(
  request: exporter_service.GarfExporterRequest,
):
  """Exports metrics continuously from API."""
  garf_exporter_service = exporter_service.GarfExporterService(
    alias=request.source,
    source_parameters=request.source_parameters,
  )
  iterations = None
  export_metrics = True
  while export_metrics:
    garf_exporter_service.generate_metrics(request, exporter)
    if request.runtime_options.expose_type == 'pushgateway':
      prometheus_client.push_to_gateway(
        gateway=request.runtime_options.address,
        job=request.runtime_options.job_name,
        registry=exporter.registry,
      )
      export_metrics = False
    await asyncio.sleep(request.runtime_options.delay_minutes * 60)
    if iterations := iterations or request.runtime_options.iterations:
      iterations -= 1
      if iterations == 0:
        export_metrics = False


async def startup_event(
  request: exporter_service.GarfExporterRequest,
):
  """Starts async task for metrics export."""
  asyncio.create_task(start_metric_generation(request))


@app.get('/health')
def health(request: fastapi.Request):
  """Defines healthcheck endpoint for GarfExporter."""
  host = request.url.hostname
  port = request.url.port
  if not healthcheck(host, port):
    raise fastapi.HTTPException(status_code=404, detail='Not updated properly')


@typer_app.command(
  context_settings={'allow_extra_args': True, 'ignore_unknown_options': True}
)
def main(
  ctx: typer.Context,
  source: Annotated[
    str,
    typer.Option(
      '-s', '--source', help='API alias', prompt='Please add API alias'
    ),
  ],
  config: Annotated[
    str,
    typer.Option(
      '-c',
      '--config',
      help='Path to configuration file',
      prompt='Please add path to config file.',
    ),
  ],
  expose_type: Annotated[
    str,
    typer.Option(help='Type of metric expose'),
  ] = 'http',
  loglevel: Annotated[
    str,
    typer.Option(help='Level of logging'),
  ] = 'INFO',
  logger: Annotated[
    str,
    typer.Option(help='Type of logging'),
  ] = 'local',
  namespace: Annotated[
    str,
    typer.Option(help='Namespace prefix for Prometheus'),
  ] = 'garf',
  host: Annotated[
    str,
    typer.Option(help='Host for exposing metrics'),
  ] = '0.0.0.0',
  port: Annotated[
    int,
    typer.Option(help='Port for exposing metrics'),
  ] = 8000,
  delay_minutes: Annotated[
    int,
    typer.Option(help='Delay in minutes between exports'),
  ] = 15,
  fetching_timeout: Annotated[
    int,
    typer.Option(help='Timeout in second for restarting stale exports'),
  ] = 120,
  iterations: Annotated[
    int,
    typer.Option(help='Stop export after N iterations'),
  ] = 0,
) -> None:
  garf_utils.init_logging(
    loglevel=loglevel,
    logger_type=logger,
    name='garf-exporter',
  )
  cli_parameters = garf_utils.ParamsParser(['macro', 'source']).parse(ctx.args)
  runtime_options = exporter_service.GarfExporterRuntimeOptions(
    expose_type=expose_type,
    host=host,
    port=port,
    namespace=namespace,
    fetching_timeout=fetching_timeout,
    iterations=iterations,
    delay_minutes=delay_minutes,
  )
  request = exporter_service.GarfExporterRequest(
    source=source,
    source_parameters=cli_parameters.get('source'),
    collectors_config=config,
    query_parameters={
      'macro': cli_parameters.get('macro', {}),
      'template': cli_parameters.get('template', {}),
    },
    runtime_options=runtime_options,
  )
  exporter.namespace = request.runtime_options.namespace

  async def start_uvicorn():
    await startup_event(request)
    config = uvicorn.Config(
      app,
      host=request.runtime_options.host,
      port=request.runtime_options.port,
      reload=True,
    )
    server = uvicorn.Server(config)
    await server.serve()

  asyncio.run(start_uvicorn())
  logging.shutdown()


if __name__ == '__main__':
  typer_app()
